import pandas as pd
from sklearn.preprocessing import RobustScaler

from isv.scripts.constants import LOSS_ATTRIBUTES, GAIN_ATTRIBUTES
from isv.config import settings


def prepare(X, cnv_type, return_train=False):
    """
    Extract relevant attributes for training and return training dataset
    together with labels, and scale the dataset - do same for validation dataset

    :param cnv_type: type of the cnv == ["loss", "gain"]
    :param X: pandas dataframe
    :param return_train: specify if transformed train data should be returned
    :return X: transformed dataframe if return_train is False. else tuple (X, X_train)
    """

    if cnv_type == 'loss':
        attributes = LOSS_ATTRIBUTES
    else:
        attributes = GAIN_ATTRIBUTES

    train_data_path = settings.data_dir + f'train_{cnv_type}.tsv.gz'
    X_train = pd.read_csv(train_data_path, compression='gzip', sep='\t')

    # Train
    X_train = X_train.loc[:, attributes]

    # evaluated data
    X_any = X.loc[:, attributes]

    # Scale
    scaler = RobustScaler()
    X_train = scaler.fit_transform(X_train)
    X_any = scaler.transform(X_any)

    if return_train:
        return X_any, X_train

    return X_any



import json
import gzip
from sklearn_json import from_json, from_dict
import xgboost as xgb


def open_model(model_path):
    """Open and return a model from json file

    :param model_path: path to the model
    """
    if model_path.endswith('gz'):
        with gzip.open(model_path, 'r') as f:
            model = f.read()
            model = json.loads(model.decode('utf-8'))
            model = from_dict(model)
        return model

    else:
        with open(model_path, 'r') as f:
            a = f.readline()

        if a.startswith('{"learner"'):
            model = xgb.Booster()
            model.load_model(model_path)
            return model

        else:
            model = from_json(model_path)
            return model



import shap

from isv.config import settings
from isv.scripts.prepare_df import prepare
from isv.scripts.open_model import open_model


def shap_values(X_raw, cnv_type):
    """Calculate SHAP values

    :param X_raw: Raw counts of genomic elements
    :param cnv_type: type of cnv
    :return: explainer object
    """
    X, X_train = prepare(X_raw, cnv_type, return_train=True)
    model = open_model(settings.model_dir + f'ISV_{cnv_type}.json')

    explainer = shap.TreeExplainer(
        model,
        X_train,
        model_output='probability')

    return explainer(X)



import pathlib


class Settings:
    def __init__(self):
        root_dir = str(pathlib.Path(__file__).parent.absolute())
        self.model_dir = root_dir + '/models/'
        self.data_dir = root_dir + '/data/'


settings = Settings()


"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

CURRENT_DIR = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (CURRENT_DIR / 'README.md').read_text(encoding='utf-8')

# Setting up
setup(
    name="isv",
    version="0.2.4",
    author="Tomas Sladecek",
    author_email="tomas.sladecek@geneton.sk",
    description='Automated Interpretation of Structural Copy Number Variants',
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.6, <4',
    install_requires=["numpy", "xgboost", "pandas", "shap"],
    keywords=['python', 'machine learning', 'copy number variation'],
    license_files=("LICENSE.txt"),
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        
        'License :: Free for non-commercial use',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate you support Python 3. These classifiers are *not*
        # checked by 'pip install'. See instead 'python_requires' below.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
    ],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
CNV specific constants and attribute lists
"""
LOSS_ATTRIBUTES = [
    'gencode_genes',
    'protein_coding',
    'pseudogenes',
    'mirna',
    'lncrna',
    'rrna',
    'snrna',
    'morbid_genes',
    'disease_associated_genes',
    'hi_genes',
    'regions_HI',
    'regulatory',
    'regulatory_enhancer',
    'regulatory_open_chromatin_region',
    'regulatory_promoter',
    'regulatory_promoter_flanking_region',
    'regulatory_ctcf_binding_site',
    'regulatory_tf_binding_site',
    'regulatory_curated'
]


GAIN_ATTRIBUTES = [
    'gencode_genes',
    'protein_coding',
    'pseudogenes',
    'mirna',
    'lncrna',
    'rrna',
    'snrna',
    'morbid_genes',
    'disease_associated_genes',
    'regions_TS',
    'regulatory',
    'regulatory_enhancer',
    'regulatory_open_chromatin_region',
    'regulatory_promoter',
    'regulatory_promoter_flanking_region',
    'regulatory_ctcf_binding_site',
    'regulatory_tf_binding_site',
    'regulatory_curated'
]


HUMAN_READABLE = {
    'length': 'CNV Length',
    'gencode_genes': 'Overlapped Gencode Elements',
    'protein_coding': 'Protein Coding Genes',
    'morbid_genes': 'Morbid Genes',
    'disease_associated_genes': 'Disease associated Genes',
    'pseudogenes': 'Pseudogenes',
    'mirna': 'Micro RNA',
    'lncrna': 'Long non-coding RNA',
    'rrna': 'Ribosomal RNA',
    'snrna': 'Small nuclear RNA',
    'hi_genes': 'Haploinsufficient Genes',
    'ts_genes': 'Triplosensitive Genes',
    'regions_HI': 'Haploinsufficient Regions',
    'regions_TS': 'Triplosensitive Regions',
    'regulatory': 'Regulatory Elements',
    'regulatory_enhancer': 'Enhancers',
    'regulatory_open_chromatin_region': 'Open Chromatin Regions',
    'regulatory_promoter': 'Promoters',
    'regulatory_promoter_flanking_region': 'Promoter Flanking Regions',
    'regulatory_ctcf_binding_site': 'CTCF Binding sites',
    'regulatory_tf_binding_site': 'TF Binding sites',
    'regulatory_curated': 'Manually Curated Regulatory Elements'
    }


DESCRIPTIONS = {
    'length' : 'Length of the candidate CNV in base-pairs',
    'gencode_genes': 'Number of Overlapped Gene Elements extracted from Gencode database',
    'protein_coding': 'Number of Protein Coding Genes (Gencode)',
    'morbid_genes': 'Number of Genes intolerant to an irregular number of copies (OMIM)',
    'disease_associated_genes': 'Number of Genes associated with some Mendelian Disease (OMIM)',
    'pseudogenes': 'Number of Pseudogenes (Gencode)',
    'mirna': 'Number of Micro RNA elements',
    'lncrna': 'Number of Long non-coding RNA elements',
    'rrna': 'Number of Ribosomal RNA elements',
    'snrna': 'Number of Small nuclear RNA elements',
    'hi_genes': 'Number Haploinsufficient Genes (score = 3) (ClinGen)',
    'ts_genes': 'Number of Triplosensitive Genes (score = 3) (ClinGen)',
    'regions_TS': 'Triplosensitive Regions (score = 3) (ClinGen)',
    'regions_HI': 'Haploinsufficient Regions (score = 3) (ClinGen)',
    'regulatory': 'Number of Regulatory Elements (NCBI Regulatory)',
    'regulatory_enhancer': 'Number of Enhancers',
    'regulatory_open_chromatin_region': 'Number of Open Chromatin Regions',
    'regulatory_promoter': 'Number of Promoters',
    'regulatory_promoter_flanking_region': 'Number of Promoter Flanking Regions',
    'regulatory_ctcf_binding_site': 'Number of CTCF Binding sites',
    'regulatory_tf_binding_site': 'Number of TF Binding sites',
    'regulatory_curated': 'Number of Manually Curated Regulatory Elements'
    }



import xgboost as xgb
import shap

from isv.config import settings
from isv.scripts.prepare_df import prepare
from isv.scripts.open_model import open_model


def predict(X_raw, cnv_type, proba=True):
    """Return model predictions for a selected dataframe
`
    :param X_raw: Raw counts of genomic elements
    :param cnv_type: type of cnv
    :param proba: return probabilities
    :return: yhat: predicted values
    """
    model = open_model(settings.model_dir + f'ISV_{cnv_type}.json')
    X = prepare(X_raw, cnv_type)

    if proba:
        if isinstance(model, xgb.core.Booster):
            X_dmat = xgb.DMatrix(X)
            yhat = model.predict(X_dmat)
        else:
            yhat = model.predict_proba(X)[:, 1]

    else:
        if isinstance(model, xgb.core.Booster):
            X_dmat = xgb.DMatrix(X)
            yhat = model.predict(X_dmat)
            yhat = (yhat > 0.5) * 1
        else:
            yhat = model.predict(X)

    return yhat